/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 10/07/23.
//

#ifndef FGGL_COMPONENTS_CORE_MATHS_COMMON_HPP
#define FGGL_COMPONENTS_CORE_MATHS_COMMON_HPP

#include <cmath>

#define GLM_ENABLE_EXPERIMENTAL
#include <glm/glm.hpp>

namespace fggl::maths {

	template<typename T>
	struct Scale {
		T start;
		T end;

		[[nodiscard]]
		constexpr bool inRange(T value) const {
			return start <= value && value <= end;
		}

		[[nodiscard]]
		constexpr T lerp(T value) const {
			assert(T(0) <= value && value <= T(1));
			return (T(1) - value) * start + value * end;
		}

		[[nodiscard]]
		constexpr T midpoint() const {
			return (end + start) / T(2);
		}

		[[nodiscard]]
		constexpr T range() const {
			return end - start;
		}

		[[nodiscard]]
		constexpr T wrap(float value) const {
			value -= start;
			return value - (range() * std::floor(value / range())) + start;
		}

		[[nodiscard]]
		constexpr T clamp(T value) const {
			const T valueOrMin = value < start ? start : value;
			return valueOrMin > end ? end : valueOrMin;
		}

		[[nodiscard]]
		constexpr T rescale(T input, Scale target) const {
			return ((input - start) * target.range()) / range();
		}

		constexpr static Scale make(T minValue, T maxValue) {
			assert( minValue < maxValue );
			return {minValue, maxValue};
		}
	};

} // namespace fggl::maths

#endif //FGGL_COMPONENTS_CORE_MATHS_COMMON_HPP
