/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 10/12/22.
//

#include "hexboard/scene.hpp"

namespace demo::hexboard {
	constexpr float SCROLL_SPEED = 0.01F;
	constexpr float HEX_SIZE = 64.0F;
	constexpr std::array<fggl::grid::IntHex, 4> ISLAND_CENTERS {{
																	{2, 3},
																	{6, 7},
																	{9, 10},
																	{6, 3}
	}};


	Scene::Scene(fggl::App &app) : GameBase(app), m_board(nullptr), m_screen(1920.F, 1080.F) {
	}

	void Scene::activate() {
		m_board = std::make_unique<fggl::grid::HexGrid>();
		m_layout = std::make_unique<fggl::grid::Layout>( fggl::grid::Orientation::make_pointy(), HEX_SIZE );
		//m_layout->m_origin = { 1920.0F * -0.5F, 1080.0F * -0.5F};

		m_selections = std::make_unique<SelectionModel>();

		const fggl::grid::TerrainType grass{
			.data = std::make_shared<fggl::grid::MaterialData>()

		};
		grass.data->name = "grass";
		grass.data->colour = {0.0F, 1.0F, 0.0};

		for (auto islandPoint : ISLAND_CENTERS){
			auto island = islandPoint.hexesInRange(2);
			for (auto &hex : island) {
				m_board->setTerrain(hex, grass);
			}
		}

		m_camera = std::make_unique<Camera2D>();
	}

	void Scene::deactivate() {
		m_board = nullptr;
		m_selections = nullptr;
		m_layout = nullptr;
	}

	void Scene::update(float delta) {
		GameBase::update(delta);

		// if the board is not set, abort
		if ( m_board == nullptr ){
			return;
		}

		// check if a button was pressed
		auto& input = this->input();
		{
			const auto mouseNdc = fggl::input::mouse_axis(input.mouse);
			const auto screenPos = ndc_to_screen(mouseNdc, m_screen);

			// calculate what the user clicked on
			auto worldPx = m_camera->unproject(screenPos);
			m_selections->hover = fggl::grid::round2( m_layout->toGrid(worldPx) );

			if (input.mouse.pressed(fggl::input::MouseButton::LEFT)) {
				m_selections->selected = m_selections->hover;
				m_camera->moveBy(mouseNdc * (m_screen * 0.5F) );
			}

			if ( input.mouse.down(fggl::input::MouseButton::RIGHT) ) {
				if (input.mouse.pressed( fggl::input::MouseButton::RIGHT )) {
					m_dragging = screenPos;
				}

				auto offset = screenPos - m_dragging.value();
				m_camera->teleportBy(offset * SCROLL_SPEED);
			} else if ( input.mouse.released(fggl::input::MouseButton::RIGHT) ) {
				m_dragging = {};
			}

		}

		m_camera->update(delta);

		// flip y, because reasons
		//auto offset = m_camera->getFocusLocation();
		//m_layout->m_origin = -offset;
	}

	void Scene::drawGrid(fggl::gfx::Paint& paint) {
		auto tiles = m_board->getAllTiles();
		for ( const auto& tile : tiles ) {
			auto terrain = m_board->getTerrain(tile);
			if ( terrain.has_value() ) {
				const auto& terrainData = terrain.value();
				m_layout->paintHex(paint, tile, terrainData.colour, m_camera->getFocusLocation());
			}
		}
	}

	void Scene::drawSelections(fggl::gfx::Paint& paint) {
		if ( m_selections == nullptr || m_dragging.has_value() ) {
			return;
		}

		if ( m_selections->selected.has_value() ) {
			m_layout->paintHex( paint, m_selections->selected.value(), fggl::gfx::colours::YELLOW, m_camera->getFocusLocation());
		}

		if ( m_selections->hover.has_value() ) {
			m_layout->paintHex( paint, m_selections->hover.value(), fggl::gfx::colours::BLANCHED_ALMOND, m_camera->getFocusLocation());
		}
	}

	void Scene::render(fggl::gfx::Graphics &gfx) {
		// if the board is not set, abort
		if ( m_board == nullptr ){
			return;
		}

		// draw the grid
		// FIXME don't hard-code the screen size
		fggl::gfx::Paint paint;
		drawGrid(paint);
		drawSelections(paint);
		gfx.draw2D(paint);
	}



} // namespace demo::hexboard