/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 19/11/22.
//

#include <stack>
#include <vector>
#include <memory>
#include <set>
#include <map>
#include <concepts>

#include "fggl/util/guid.hpp"

namespace fggl::assets {

	using AssetName = util::OpaqueName<uint64_t, struct AssetRefStruct>;

	/**
	 * The base class representing an asset.
	 *
	 * This can be combined with the templated version.
	 */
	class Asset {
		public:
			Asset(AssetName name) : m_name(name) {}
			virtual ~Asset() = default;

			inline void release() {
				releaseImpl();
			}

			bool operator==(const Asset& asset) const {
				return m_name == asset.m_name;
			}

			bool operator!=(const Asset& asset) const {
				return m_name != asset.m_name;
			}

		private:
			AssetName m_name;
			virtual void releaseImpl() = 0;
	};

	/**
	 * Wrapper for types that cannot extend Asset themselves.
	 *
	 * @tparam T the asset to wrap
	 */
	template<typename T>
	class AssetBox : public Asset {
		public:
			AssetBox(T* ptr) : m_ptr(ptr) {}

			~AssetBox() override {
				release();
			}

			T* ptr() {
				return m_ptr;
			}

		private:
			T* m_ptr;

			void releaseImpl() override {
				delete m_ptr;
				m_ptr = nullptr;
			}

	};

	/**
	 * Asset Library.
	 *
	 * The currently usable set of assets loaded into the engine.
	 */
	class AssetLibrary {
		public:
			void load(AssetName name);
			void unload(AssetName name);

			template<typename T>
			requires std::derived_from<T, Asset>
			void store(AssetName name, T* ptr) {
				m_assets[name] = std::make_unique<T>(ptr);
			}

			template<typename T>
			requires std::derived_from<T, Asset>
			void get(AssetName name) const {
				Asset* asset = m_assets.at(name).get();
				return dynamic_cast<T>(asset);
			}

			inline Asset* get(AssetName name) const {
				return m_assets.at( name ).get();
			}

		private:
			std::map<AssetName, std::unique_ptr<Asset>> m_assets;

	};

	struct AssetRecord {
		AssetName id;
		std::vector<AssetName> dependencies;

		bool hasDepends(AssetName name) const;
		void addDepend(AssetName name);
	};

	class AssetGraph {
		public:
			AssetGraph(AssetLibrary* loader);

			/**
			 *
			 *
			 * @param name
			 */
			void require(AssetName& name) {
				if ( m_loaded.find(name) != m_loaded.end() ) {
					return;
				}
				m_required.push(name);
			}

			/**
			 *
			 *
			 * @return
			 */
			 [[nodiscard]]
			inline bool isLoadComplete() const {
				return m_required.empty();
			}

			/**
			 *
			 *
			 * @param name
			 * @return
			 */
			 [[nodiscard]]
			inline bool isLoaded(AssetName name) const {
				return m_loaded.find(name) != m_loaded.end();
			}

			/**
			 *
			 * @param name
			 * @param description
			 */
			void addDependency(AssetName& name, AssetName& description) {
				auto& assets = m_assets.at(name);
				assets.addDepend(description);
			}

			/**
			 *
			 * @param name
			 * @param dependency
			 * @return
			 */
			[[nodiscard]]
			bool hasDependency(AssetName& name , AssetName& dependency) const {
				try {
					auto &assets = m_assets.at(name);
					return assets.hasDepends(dependency);
				} catch ( std::out_of_range& ) {
					return false;
				}
			}

			/**
			 *
			 */
			void process() {
				auto asset = m_required.top();
				m_loader->load( asset );
				m_required.pop();
			}

			/**
			 *
			 */
			void finishLoading() {
				while ( !m_loaded.empty() ) {
					process();
				}
			}

		private:
			std::stack<AssetName> m_required;
			std::set<AssetName> m_loaded;
			std::map<AssetName, AssetRecord> m_assets;
			AssetLibrary* m_loader;
	};

} // namespace fggl::assets