/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include <fggl/gfx/input.hpp>

using fggl::gfx::Input;

Input::Input() : m_mouse_curr(), m_mouse_last(), m_joysticks(), m_joydata(), m_pad_curr(), m_pad_last() {
	clear();
}

void Input::clear() {
	// reset mouse data
	m_mouse_curr.cursor[0] = 0.0;
	m_mouse_curr.cursor[1] = 0.0;
	m_mouse_curr.scroll[0] = 0.0;
	m_mouse_curr.scroll[1] = 0.0;
	m_mouse_curr.buttons = 0x00;
	m_mouse_last = m_mouse_curr;

	// reset pad data
	for (auto &pad : m_pad_curr) {
		pad.present = false;
		pad.buttons.reset();
		for (float &axe : pad.axes) {
			axe = 0.0F;
		}
	}
	m_pad_last = m_pad_curr;
}

void Input::frame() {
	m_mouse_last = m_mouse_curr;
	m_mouse_curr.scroll[0] = 0.0;
	m_mouse_curr.scroll[1] = 0.0;

	m_pad_last = m_pad_curr;
}

void Input::mousePos(double x, double y) {
	m_mouse_curr.cursor[0] = x;
	m_mouse_curr.cursor[1] = y;
}

auto Input::cursorDeltaX() const -> double {
	return m_mouse_last.cursor[0] - m_mouse_curr.cursor[0];
}

auto Input::cursorDeltaY() const -> double {
	return m_mouse_last.cursor[1] - m_mouse_curr.cursor[1];
}

auto Input::mousePos() const -> const double * {
	return m_mouse_curr.scroll.data();
}

void Input::mouseScroll(double deltaX, double deltaY) {
	m_mouse_curr.scroll[0] = deltaX;
	m_mouse_curr.scroll[1] = deltaY;
}

auto Input::mouseScroll() const -> const double * {
	return m_mouse_curr.scroll.data();
}

auto Input::scrollDeltaX() const -> double {
	return m_mouse_curr.scroll[0];
}

auto Input::scrollDeltaY() const -> double {
	return m_mouse_curr.scroll[1];
}

void Input::mouseBtn(const MouseButton btn, bool state) {
	if (state) {
		m_mouse_curr.buttons |= btn;
	} else {
		m_mouse_curr.buttons &= ~btn;
	}
}

auto Input::mouseDown(const MouseButton btn) const -> bool {
	return m_mouse_curr.buttons & btn;
}

auto Input::mousePressed(const MouseButton btn) const -> bool {
	return (m_mouse_curr.buttons & btn) && !(m_mouse_last.buttons & btn);
}

auto Input::mouseReleased(const MouseButton btn) const -> bool {
	return !(m_mouse_curr.buttons & btn) && (m_mouse_last.buttons & btn);
}

void Input::joystickConnect(int id, Joystick &data) {
	// joystick data is polled, so we need to call this every frame >.<
	m_joysticks[id] = true;
	m_joydata[id] = data;
}

void Input::joystickDisconnect(int id) {
	// reset to empty joystick
	m_joysticks[id] = false;
	m_joydata[id] = Joystick();
}

auto Input::hasJoystick(int id) const -> bool {
	return m_joysticks[id];
}

auto Input::joystick(int id) const -> const fggl::gfx::Joystick & {
	return m_joydata[id];
}

void Input::padState(int id, const PadState &state) {
	m_pad_curr[id] = state;
}

auto Input::padDown(int id, PadButton btn) -> bool {
	return m_pad_curr[id].buttons[(int) btn];
}

auto Input::padPressed(int id, PadButton btn) -> bool {
	return m_pad_curr[id].buttons[(int) btn] && !m_pad_last[id].buttons[(int) btn];
}

auto Input::padReleased(int id, PadButton btn) -> bool {
	return !m_pad_curr[id].buttons[(int) btn] && m_pad_last[id].buttons[(int) btn];
}

auto Input::padAxis(int id, PadAxis axis) -> float {
	return m_pad_curr[id].axes[(int) axis];
}

auto Input::padAxisDelta(int id, PadAxis axis) -> float {
	return m_pad_last[id].axes[(int) axis] - m_pad_curr[id].axes[(int) axis];
}
