/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include "fggl/gfx/ogl4/meshes.hpp"
#include "fggl/gfx/phong.hpp"

#include "fggl/data/texture.hpp"

//
// Created by webpigeon on 22/10/22.
//

namespace fggl::gfx::ogl4 {

	static auto setup_array_buffer(std::shared_ptr<ogl::VertexArray> &vao,
															  const std::vector<mesh::Vertex3D> &data) -> std::shared_ptr<ogl::ArrayBuffer> {
		// upload the data to the GPU
		auto buff = std::make_shared<ogl::ArrayBuffer>();
		buff->write(data.size() * sizeof(mesh::Vertex3D), data.data(), ogl::BufUsage::STATIC_DRAW);

		// set up the vertex attributes
		auto posAttr = ogl::attribute<mesh::Vertex3D, math::vec3>(offsetof(mesh::Vertex3D, position));
		auto normalAttr = ogl::attribute<mesh::Vertex3D, math::vec3>(offsetof(mesh::Vertex3D, normal));
		auto colAttr = ogl::attribute<mesh::Vertex3D, math::vec3>(offsetof(mesh::Vertex3D, colour));
		auto texAttr = ogl::attribute<mesh::Vertex3D, math::vec2>(offsetof(mesh::Vertex3D, texPos));

		vao->setAttribute(*buff, 0, posAttr);
		vao->setAttribute(*buff, 1, normalAttr);
		vao->setAttribute(*buff, 2, colAttr);
		vao->setAttribute(*buff, 3, texAttr);

		return buff;
	}

	static auto setup_index_buffer(const std::vector<uint32_t>& data) -> std::shared_ptr<ogl::ElementBuffer> {
		auto elementBuffer = std::make_shared<ogl::ElementBuffer>();
		elementBuffer->write(data.size() * sizeof(uint32_t),
							 data.data(), ogl::BufUsage::STATIC_DRAW);
		return elementBuffer;
	}

	void setup_material(const std::shared_ptr<ogl::Shader>& shader, const PhongMaterial* material) {
		if ( !shader->hasUniform("material.ambient") ) {
			return;
		}

		// setup material block
		shader->setUniformF(shader->uniform("material.emission"), material->emission);
		shader->setUniformF(shader->uniform("material.ambient"), material->ambient);
		shader->setUniformF(shader->uniform("material.diffuse"), material->diffuse);
		shader->setUniformF(shader->uniform("material.specular"), material->specular);
		//shader->setUniformF(shader->uniform("material.shininess"), material->shininess);
	}

	void setup_material(const std::shared_ptr<ogl::Shader>& shader, const Material* material) {
		if ( shader->hasUniform("material.diffuse") ) {
			// setup material block
//			shader->setUniformF(shader->uniform("material.emission"), material->emission);
//			shader->setUniformF(shader->uniform("material.ambient"), material->ambient);
			shader->setUniformF(shader->uniform("material.diffuse"), material->m_diffCol);
			shader->setUniformF(shader->uniform("material.specular"), material->m_specCol);
		}


		// setup diffuse texture
		if (material->m_diffuse != nullptr) {
			material->m_diffuse->bind(0);
			if (shader->hasUniform("diffuseTexture")) {
				shader->setUniformI(shader->uniform("diffuseTexture"), 0);
			}
		}

		// setup specular texture
		if (material->m_specular != nullptr) {
			material->m_specular->bind(1);
			if (shader->hasUniform("specularTexture")) {
				shader->setUniformI(shader->uniform("specularTexture"), 1);
			}
		}
		//shader->setUniformF(shader->uniform("material.shininess"), material->shininess);
	}


	void MeshData::draw(std::shared_ptr<ogl::Shader> shader) const {
		vao->bind();
		vertexData->bind();

		if ( material != nullptr ) {
			setup_material(shader, material);
		} else {
			debug::info("no material is active, cannot bind textures!");
		}

		if ( drawInfo.restartIndex != ogl::NO_RESTART_IDX) {
			glEnable(GL_PRIMITIVE_RESTART);
			glPrimitiveRestartIndex( drawInfo.restartIndex );
		}

		vao->drawElements(*elements, drawInfo.mode, elementCount);

		if ( drawInfo.restartIndex != ogl::NO_RESTART_IDX ) {
			glDisable(GL_PRIMITIVE_RESTART);
		}
	}

	void StaticMultiMesh::draw() const {
		for ( const auto& mesh : meshes ) {
			mesh.draw(pipeline);
		}
	}

	void setup_lighting(const std::shared_ptr<ogl::Shader>& shader, const DirectionalLight* light) {
		assert( light != nullptr );
		if ( !shader->hasUniform("light.direction") ) {
			fggl::debug::warning("asked for directional lighting, but shader does not support!");
			return;
		}

		shader->setUniformF( shader->uniform("light.direction"), light->position);
		shader->setUniformF( shader->uniform("light.ambient"), light->ambient);
		shader->setUniformF( shader->uniform("light.diffuse"), light->diffuse);
		shader->setUniformF( shader->uniform("light.specular"), light->specular);
	}

	void setup_lighting(const std::shared_ptr<ogl::Shader>& shader, const math::mat4& viewMatrix, const math::Transform& camTransform, const math::Transform& transform, math::vec3 lightPos) {
		if (shader->hasUniform("lightPos")) {
			shader->setUniformF(shader->uniform("lightPos"), lightPos);
		}

		if ( !shader->hasUniform("lights[0].isEnabled") ) {
			return;
		}

		bool local = true;

		shader->setUniformI(shader->uniform("lights[0].isEnabled"), 1);
		shader->setUniformI(shader->uniform("lights[0].isLocal"), local);
		shader->setUniformI(shader->uniform("lights[0].isSpot"), 0);

		shader->setUniformF(shader->uniform("lights[0].constantAttenuation"), 5.0F);
		shader->setUniformF(shader->uniform("lights[0].linearAttenuation"), 0.0F);
		shader->setUniformF(shader->uniform("lights[0].quadraticAttenuation"), 0.0F);

		shader->setUniformF(shader->uniform("Strength"), 0.6F);

		if (!local) {
			lightPos = glm::normalize(lightPos);
			auto viewDir = glm::normalize(camTransform.origin() - transform.origin());
			auto halfVector = glm::normalize(lightPos + viewDir);
			shader->setUniformF(shader->uniform("lights[0].halfVector"), halfVector);
			shader->setUniformF(shader->uniform("EyeDirection"), viewDir);
			shader->setUniformF(shader->uniform("lights[0].position"), lightPos);
		} else {
			auto camModelView = (viewMatrix * camTransform.model() * math::vec4(0.0F, 0.0F, 0.0F, 1.0F));
			auto modelModelView = (viewMatrix * transform.model() * math::vec4(0.0F, 0.0F, 0.0F, 1.0F));
			math::vec3 viewDir = glm::normalize(camModelView - modelModelView);
			shader->setUniformF(shader->uniform("EyeDirection"), viewDir);

			shader->setUniformF(shader->uniform("lights[0].position"),
								math::vec3(viewMatrix * math::vec4(lightPos, 1.0F)));
		}

		shader->setUniformF(shader->uniform("lights[0].ambient"), {0.0F, 0.5F, 0.0F});
		shader->setUniformF(shader->uniform("lights[0].colour"), {0.5F, 0.5F, 0.5F});
	}

	static auto upload_texture( assets::AssetID name, assets::AssetManager* manager ) -> ogl::Texture* {
		debug::info("loading texture: {}", name.get());

		auto uploadedTex = assets::make_asset_id_rt("ogl", std::to_string(name.get()) );
		auto* texture = manager->get<ogl::Texture>( uploadedTex );
		if ( texture != nullptr ) {
			return texture;
		}

		// get the texture data we plan to load
		auto *textureData = manager->get<data::Texture2D>(name);

		// create a texture
		texture = new ogl::Texture(ogl::TextureType::Tex2D);
		texture->setData( ogl::InternalImageFormat::RedGreenBlue, textureData);

		return manager->set(uploadedTex, texture);
	}

	static auto get_fallback_material(assets::AssetManager* manager) -> Material* {
		auto* fallback = manager->get<Material>(FALLBACK_MAT);
		if ( fallback != nullptr ) {
			return fallback;
		}

		auto* mat = new Material();
		mat->m_diffCol = FALLBACK_COLOUR;
		mat->m_specCol= FALLBACK_COLOUR;

		mat->m_diffuse = manager->get<ogl::Texture>(FALLBACK_TEX);
		mat->m_specular = manager->get<ogl::Texture>(FALLBACK_TEX);
		mat->m_normals = manager->get<ogl::Texture>(SOLID_TEX);
		return manager->set(FALLBACK_MAT, mat);
	}

	static auto upload_material( assets::AssetID name, assets::AssetManager* manager ) -> Material* {
		auto* meshMaterial = manager->get<mesh::Material>(name);
		if ( meshMaterial == nullptr ){
			debug::error("attempted to load material {}, but did not exist!", name.get());
			return get_fallback_material(manager);
		}

		auto uploadedMat = assets::make_asset_id_rt("ogl", std::to_string(name.get()) );
		auto* material = manager->get<Material>( uploadedMat );
		if ( material != nullptr ) {
			return material;
		}

		material = new Material();
		material->m_diffCol = meshMaterial->diffuse;
		material->m_specCol = meshMaterial->specular;

		// do we have a diffuse texture?
		if ( !meshMaterial->diffuseTextures.empty() ) {
			material->m_diffuse = upload_texture(meshMaterial->getPrimaryDiffuse(), manager);
		} else {
			material->m_diffuse = manager->get<ogl::Texture>( ogl4::SOLID_TEX );
		}

		// do we have a normal Texture?
		if ( !meshMaterial->normalTextures.empty() ) {
			material->m_normals = upload_texture(meshMaterial->getPrimaryNormals(), manager);
		} else {
			material->m_normals = manager->get<ogl::Texture>( ogl4::SOLID_TEX );
		}

		// do we have a specular texture?
		if ( !meshMaterial->specularTextures.empty() ) {
			material->m_specular = upload_texture(meshMaterial->getPrimarySpecular(), manager);
		} else {
			material->m_specular = manager->get<ogl::Texture>( ogl4::SOLID_TEX );
		}

		return manager->set( uploadedMat, material);
	}

	auto upload_mesh(const mesh::Mesh3D& rawMesh, assets::AssetManager* manager) -> MeshData {
		auto vao = std::make_shared<ogl::VertexArray>();
		return {
			.vao = vao,
			.elements = setup_index_buffer( rawMesh.indices ),
			.vertexData = setup_array_buffer(vao, rawMesh.data ),
			.elementCount = rawMesh.indices.size(),
			.drawInfo = { .mode = ogl::Primitive::TRIANGLE },
			.material = upload_material( rawMesh.material, manager )
		};
	}

	auto upload_multi_mesh(const mesh::MultiMesh3D& rawMesh, assets::AssetManager* manager) -> std::vector<MeshData> {
		std::vector<MeshData> gpuMeshes;
		for (const auto& mesh : rawMesh.meshes) {
			gpuMeshes.push_back( upload_mesh(mesh, manager) );
		}
		return gpuMeshes;
	}

}
