/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 18/04/22.
//

#include "fggl/gfx/ogl4/models.hpp"
#include "fggl/gfx/ogl4/meshes.hpp"

#include "fggl/data/heightmap.hpp"
#include "fggl/data/texture.hpp"

#include "fggl/debug/logging.hpp"
#include "fggl/gfx/camera.hpp"
#include "fggl/gfx/phong.hpp"

namespace fggl::gfx::ogl4 {

	static auto setup_array_buffer(std::shared_ptr<ogl::VertexArray> &vao,
															  const std::vector<data::Vertex> &data) -> std::shared_ptr<ogl::ArrayBuffer> {
		auto buff = std::make_shared<ogl::ArrayBuffer>();
		buff->write(data.size() * sizeof(data::Vertex), data.data(), ogl::BufUsage::STATIC_DRAW);

		// set up the vertex attributes
		auto posAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, posititon));
		auto normalAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, normal));
		auto colAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, colour));

		vao->setAttribute(*buff, 0, posAttr);
		vao->setAttribute(*buff, 1, normalAttr);
		vao->setAttribute(*buff, 2, colAttr);
		return buff;
	}


	static auto setup_index_buffer(std::shared_ptr<ogl::VertexArray> &vao,
																const std::vector<uint32_t> &data) -> std::shared_ptr<ogl::ElementBuffer> {
		auto elementBuffer = std::make_shared<ogl::ElementBuffer>();
		elementBuffer->write(data.size() * sizeof(uint32_t),
							 data.data(), ogl::BufUsage::STATIC_DRAW);
		return elementBuffer;
	}

	static void setup_component(StaticModel &modelComp, std::shared_ptr<ogl::Shader> &shader, data::Mesh &mesh) {
		auto vao = std::make_shared<ogl::VertexArray>();
		auto meshBuffer = setup_array_buffer(vao, mesh.vertexList());
		auto elementBuffer = setup_index_buffer(vao, mesh.indexList());

		// set up the element attributes
		modelComp.vao = vao;
		modelComp.vertexData = meshBuffer;
		modelComp.elements = elementBuffer;
		modelComp.pipeline = shader;
		modelComp.elementCount = mesh.indexCount();
		modelComp.drawType = ogl::Primitive::TRIANGLE;
	}

	auto StaticModelRenderer::uploadMesh(assets::AssetID guid, const data::Mesh &mesh, bool allowCache) -> StaticModel* {
		assert( m_assets != nullptr );

		// if the asset has already been uploaded, we don't need to do anything
		if ( allowCache && m_assets->has(guid) ) {
			m_assets->require(guid);
			return m_assets->get<StaticModel>(guid);
		}

		// the asset does not exist, we need to upload it
		auto* modelAsset = new StaticModel();
		modelAsset->vao = std::make_shared<ogl::VertexArray>();
		modelAsset->vertexData = setup_array_buffer(modelAsset->vao, mesh.vertexList());
		modelAsset->elements = setup_index_buffer(modelAsset->vao, mesh.indexList());
		modelAsset->elementCount = mesh.indexCount();
		modelAsset->drawType = ogl::Primitive::TRIANGLE;

		// if caching is enabled, then use the cache
		if ( allowCache ) {
			m_assets->set(guid, modelAsset);
		}

		return modelAsset;
	}

	/*MeshData* StaticModelRenderer::uploadMesh(assets::AssetGUID guid, const mesh::Mesh3D &mesh, bool allowCache) {
		assert( m_assets != nullptr );

		// if the asset has already been uploaded, we don't need to do anything
		if ( allowCache && m_assets->has(guid) ) {
			m_assets->require(guid);
			return m_assets->get<MeshData>(guid);
		}

		// the asset does not exist, we need to upload it
		auto* modelAsset = new MeshData();
		modelAsset->vao = std::make_shared<ogl::VertexArray>();
		modelAsset->vertexData = setupArrayBuffer(modelAsset->vao, mesh.data);
		modelAsset->elements = setupIndexBuffer(modelAsset->vao, mesh.indices);
		modelAsset->elementCount = mesh.indices.size();
		modelAsset->drawInfo.mode = ogl::Primitive::TRIANGLE;

		// if caching is enabled, then use the cache
		if ( allowCache ) {
			m_assets->set(guid, modelAsset);
		}

		return modelAsset;
	}*/

	auto StaticModelRenderer::uploadMesh2(const assets::AssetID& meshName, const data::Mesh &mesh) -> StaticModelGPU* {
		assert( m_assets != nullptr );

		if ( m_assets->has(meshName) ) {
			// we've already uploaded it...
			return m_assets->get<StaticModelGPU>(meshName);
		}

		auto* modelAsset = new StaticModelGPU();
		modelAsset->vao = std::make_shared<ogl::VertexArray>();
		modelAsset->vertices = setup_array_buffer(modelAsset->vao, mesh.vertexList());
		modelAsset->elements = setup_index_buffer(modelAsset->vao, mesh.indexList());
		modelAsset->elementCount = mesh.indexCount();
		modelAsset->drawType = ogl::Primitive::TRIANGLE;

		return m_assets->set(meshName, modelAsset);
	}

	#ifdef FGGL_ALLOW_DEFERRED_UPLOAD
		static void setup_meshes(entity::EntityManager& world, ShaderCache* shaders, assets::AssetManager* manager) {
			auto entsWithModels = world.find<mesh::StaticMesh3D>();
			for ( const auto& mesher : entsWithModels ) {

				// check if this entity already has a mesh...
				const auto* currModel = world.tryGet<StaticMesh>(mesher);
				if ( currModel != nullptr ) {
					continue;
				}

				// figure out the requirements
				auto& rawMesh = world.get<mesh::StaticMesh3D>(mesher);
				if ( rawMesh.guid != util::make_guid("__NO_CACHE__") ){
					// TODO support fetching of loaded meshes
					debug::warning("multiple entities sharing a mesh has not been implemented yet...");
				}

				auto& entityMesh = world.add<StaticMesh>(mesher);
				entityMesh.pipeline = shaders->get( rawMesh.pipeline );
				entityMesh.mesh = upload_mesh( rawMesh.mesh, manager );

			}
		}

		static void setup_multi_meshes(entity::EntityManager& world, ShaderCache* shaders, assets::AssetManager* manager) {
			auto entsWithModels = world.find<mesh::StaticMultiMesh3D>();
			for ( const auto& mesher : entsWithModels ) {

				// check if this entity already has a mesh...
				const auto* currModel = world.tryGet<StaticMultiMesh>(mesher);
				if ( currModel != nullptr ) {
					continue;
				}

				// figure out the requirements
				auto& multiMesh = world.get<mesh::StaticMultiMesh3D>(mesher);
				if ( multiMesh.guid != util::make_guid("__NO_CACHE__") ){
					// TODO support fetching of loaded meshes
					debug::warning("multiple entities sharing a mesh has not been implemented yet...");
				}

				auto& entityMesh = world.add<StaticMultiMesh>(mesher);
				entityMesh.pipeline = shaders->get( multiMesh.pipeline );
				entityMesh.meshes = upload_multi_mesh(multiMesh.mesh, manager);
				//entityMesh.material = setup_material( multiMesh.mesh.materials[0], manager);
			}
		}

		void StaticModelRenderer::resolveModels(entity::EntityManager &world) {
			// new mesh formats
			setup_meshes(world, m_shaders, m_assets);
			setup_multi_meshes(world, m_shaders, m_assets);

			// terrain
			auto terrain = world.find<data::HeightMap>();
			for (const auto &renderable : terrain) {
				auto *currModel = world.tryGet<StaticModel>(renderable);
				if (currModel != nullptr) {
					continue;
				}

				auto &heightmap = world.get<data::HeightMap>(renderable);
				data::Mesh heightMapMesh{};
				data::generateHeightMesh(heightmap, heightMapMesh);

				auto &modelComp = world.add<StaticModel>(renderable);
				setup_component(modelComp, m_phong, heightMapMesh);

				// we know this is a triangle strip with a restart vertex...
				// FIXME the model should be telling us this...
				modelComp.drawType = ogl::Primitive::TRIANGLE_STRIP;
				modelComp.restartIndex = heightMapMesh.restartVertex;

				// no active model, we need to resolve/load one.
				debug::info("looks like {} needs a static mesh", (uint64_t)renderable);
			}
		}
	#endif


	void StaticModelRenderer::renderModelsForward(const entity::EntityManager &world, bool debugMode) {

		// fetch cameras we will need to render with
		auto cameras = world.find<gfx::Camera>();

		// if there are no cameras, we can't do anything...
		if (cameras.empty()) {
			debug::warning("asked to render static models, but there were no cameras");
			return;
		}

		// perform a rendering pass for each camera (will usually only be one...)
		for (const auto &cameraEnt : cameras) {
			//TODO should be clipping this to only visible objects
			//forward_camera_pass(cameraEnt, world);

			forward_pass<ogl4::StaticMesh>(cameraEnt, world, m_assets);
			forward_pass<ogl4::StaticMultiMesh>(cameraEnt, world, m_assets);

			// enable rendering normals
			if ( debugMode ) {
				auto normalShader = m_shaders->get("normals");
				forward_pass_normals<StaticMesh>(cameraEnt, world, normalShader);
				forward_pass_normals<StaticMultiMesh>(cameraEnt, world, normalShader);
				//forward_normal_pass(cameraEnt, world, normalShader);
			}
		}
	}

}
