/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 24/07/22.
//

#include "fggl/gfx/ogl4/module.hpp"
#include "fggl/gfx/phong.hpp"
#include "fggl/mesh/components.hpp"
#include "fggl/data/procedural.hpp"

#include "fggl/assets/loader.hpp"
#include <string>

namespace fggl::gfx {

	constexpr uint32_t DEFAULT_STACKS = 16;
	constexpr uint32_t DEFAULT_SLICES = 16;
	constexpr const char *SHAPE_SPHERE{"sphere"};
	constexpr const char *SHAPE_BOX{"box"};

	namespace {
		void process_shape(const YAML::Node &node, mesh::Mesh3D &mesh) {
			auto transform = data::OFFSET_NONE;

			auto offset = node["offset"].as<math::vec3>(math::VEC3_ZERO);
			transform = glm::translate(transform, offset);

			auto scale = node["scale"].as<math::vec3>(math::VEC3_ONES);
			transform = glm::scale(transform, scale);
			debug::debug("scale: {}, {}, {}", scale.x, scale.y, scale.z);

			// now the shape itself
			auto type = node["type"].as<std::string>();
			if (type == SHAPE_BOX) {
				data::make_cube(mesh, transform);
			} else if (type == SHAPE_SPHERE) {
				auto stacks = node["stacks"].as<uint32_t>(DEFAULT_STACKS);
				auto slices = node["slices"].as<uint32_t>(DEFAULT_SLICES);
				data::make_sphere(mesh, transform, stacks, slices);
			} else {
				debug::log(debug::Level::warning, "unknown shape type requested: {}", type);
			}
		}
	}


	void attach_mesh(const entity::ComponentSpec &spec, entity::EntityManager &manager, const entity::EntityID &id, services::Generator *services) {

		// check for the asset service
		auto assetService = services->getLazy<assets::AssetManager>();
		auto assetLoader = services->getLazy<assets::Loader>();

		// asset is a procedural mesh description
		if (spec.has("shape")) {
			auto pipeline = spec.get<std::string>("pipeline", "");

			// check if we had previously loaded this asset
			/*const auto shapeName = spec.get<std::string>("shape_id", "");
			if ( !shapeName.empty() ) {
				meshAsset = assetService->get<data::Mesh>(shapeName);
			}*/

			// we've not loaded this before - generate mesh
			if ( true ) {

				// procedural meshes are build as static meshes first
				if (spec.config["shape"].IsSequence()) {
					mesh::MultiMesh3D* multiMesh = new mesh::MultiMesh3D();

					for (const auto &node : spec.config["shape"]) {
						auto* meshAsset = new mesh::Mesh3D();
						process_shape(node, *meshAsset);
						multiMesh->meshes.push_back(*meshAsset);
						delete meshAsset;
					}

					#ifdef FGGL_ALLOW_DEFERRED_UPLOAD
						// the graphics stack can detect static meshes without a rendering proxy at runtime and fix it but this
						// requires loading the whole model into the ECS (and should be removed in the future). instead we should
						// be triggering the upload at this point (if needed).
						auto &entityMesh = manager.add<mesh::StaticMultiMesh3D>(id);
						entityMesh.mesh = *multiMesh;
						entityMesh.pipeline = pipeline;
						debug::warning("HACKY: Triggered proc mesh - using deferred upload");
					#endif

				} else {
					auto* meshAsset = new mesh::Mesh3D();
					process_shape(spec.config["shape"], *meshAsset);

					#ifdef FGGL_ALLOW_DEFERRED_UPLOAD
						// the graphics stack can detect static meshes without a rendering proxy at runtime and fix it but this
						// requires loading the whole model into the ECS (and should be removed in the future). instead we should
						// be triggering the upload at this point (if needed).
						auto &entityMesh = manager.add<mesh::StaticMesh3D>(id);
						entityMesh.mesh = *meshAsset;
						entityMesh.pipeline = pipeline;
						debug::warning("HACKY: Triggered proc mesh - using deferred upload");
					#endif
				}

				//assetService->set(shapeName, meshTmp);
			}

			// TODO we need to trigger an upload to the GPU (somehow)

			return;
		}

		// asset is a model from the resource cache
		if ( spec.has("model") ) {
			// figure out what model we want
			auto assetStr = spec.get<std::string>("model", "");
			auto assetId = assets::asset_from_user( assetStr );

			auto* asset = assetService->get<ogl4::StaticModel>(assetId);

			if ( asset == nullptr ) {
				// the asset is not loaded/does not exist
				debug::error("requested model {} but it was not loaded.", assetStr);
				return;
			}

			// copy the asset to the model
			auto& model = manager.add<ogl4::StaticModel>(id);
			model = *asset;
		}
	}

	void attach_material(const entity::ComponentSpec &spec,
						 entity::EntityManager &manager,
						 const entity::EntityID &id,
						 modules::Services* /*services*/ ) {
		auto &mat = manager.add<gfx::PhongMaterial>(id);
		mat.ambient = spec.get<math::vec3>("ambient", gfx::DEFAULT_AMBIENT);
		mat.diffuse = spec.get<math::vec3>("diffuse", gfx::DEFAULT_DIFFUSE);
		mat.specular = spec.get<math::vec3>("ambient", gfx::DEFAULT_SPECULAR);
		mat.shininess = spec.get<float>("ambient", gfx::DEFAULT_SHININESS);
	}

	void attach_light_directional(const entity::ComponentSpec &spec, entity::EntityManager &manager, const entity::EntityID &id, services::Generator* /*services*/) {
		auto &light = manager.add<gfx::DirectionalLight>(id);
		light.position = spec.get<math::vec3>("direction", -math::UP);
		light.ambient = spec.get<math::vec3>("ambient", gfx::colours::WHITE);
		light.specular = spec.get<math::vec3>("specular", gfx::colours::WHITE);
		light.diffuse = spec.get<math::vec3>("diffuse", gfx::colours::WHITE);
	}

	void attach_light_point(const entity::ComponentSpec &spec, entity::EntityManager &manager, const entity::EntityID &id, services::Generator* /*services*/) {
		auto &light = manager.add<gfx::PointLight>(id);
		light.position = spec.get<math::vec3>("position", math::VEC3_ZERO);

		light.constant = spec.get<float>("constant", 1.0F);
		light.linear = spec.get<float>("linear", 0.0014F);
		light.quadratic = spec.get<float>("quadratic", 0.000007F);
	}


	void OpenGL4::wireServices(fggl::services::ModuleBinder &binder) {
		binder.bind<WindowGraphics>().to<ogl4::WindowGraphics, filesystem::Locator, filesystem::Reader, gui::FontLibrary, assets::AssetManager>();
	}

	services::Module::ServiceList OpenGL4::getProvides() const {
		return { MODULE_ID, API::MODULE_ID };
	}

	void OpenGL4::setup(services::Generator *generator) {
		auto entityFac = generator->getLazy<entity::EntityFactory>();
		entityFac->bind(data::StaticMesh::guid, attach_mesh);
		entityFac->bind(mesh::StaticMesh3D::guid, attach_mesh);
		entityFac->bind(mesh::StaticMultiMesh3D::guid, attach_mesh);

		// material
		entityFac->bind(gfx::PhongMaterial::guid, attach_material);

		// lighting
		entityFac->bind(gfx::DirectionalLight::guid, attach_light_directional);
		entityFac->bind(gfx::PointLight::guid, attach_light_point);
	}

} // namespace fggl::gfX::ogl4