/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 12/07/23.
//

#include "fggl/modules/manager.hpp"

namespace fggl::modules {

	bool Manager::buildGraph() {

		for ( auto& api : m_serviceProviders ) {
			std::cerr << api.first.get() << " is provided by " << api.second.get() << std::endl;
		}

		// resolve links between modules
		for (auto &moduleItr : m_modules) {
			debug::trace("Generating vertex for {}", moduleItr.first.get());

			// add us to the module graph
			m_dependencies.addVertex( moduleItr.first );

			// wire us to the contracts and modules we depend on
			for (auto &service : moduleItr.second->getDepends()) {

				auto provider = m_serviceProviders.find(service);
				if (provider == m_serviceProviders.end()) {
					debug::log(debug::Level::warning,
							   "{} depends on {}, but nothing provides it",
							   moduleItr.first.get(),
							   service.get());
					// nothing can provide the service requested, setup is invalid.
					return false;
				}

				m_dependencies.addEdge(moduleItr.first, provider->second);
			}
		}
		return true;
	}

	bool Manager::resolve() {
		assert( !m_locked );
		if (!buildGraph()) {
			fggl::debug::error("Failed to build module dependency graph, your configuration is broken!");
			return false;
		}

		std::queue<ModuleIdentifier> stack;
		m_dependencies.getOrderRev(stack);

		services::ModuleBinder binder(&m_services);
		while (!stack.empty()) {
			auto nextToInit = stack.front();
			debug::log(debug::Level::info, "Initializing {}", nextToInit.get());

			auto &module = m_modules.at(nextToInit);
			module->setup( &m_services );
			module->wireServices( binder );
			stack.pop();
		}

		debug::log(debug::Level::info, "engine boot complete");
		m_locked = true;
		return true;
	}

} // namespace fggl::modules