/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_AUDIO_AUDIO_HPP
#define FGGL_AUDIO_AUDIO_HPP

#include <string>

#include "fggl/data/storage.hpp"
#include "fggl/modules/module.hpp"
	#include "fggl/services/module.hpp"
#include "fggl/assets/module.hpp"
#include "fggl/assets/packed/module.hpp"

//! backend independent audio interface
namespace fggl::audio {

	/**
	 * AudioClip is bit of audio loaded into memory.
	 *
	 * If the sampleCount is -1, the clip is invalid.
	 */
	 template<typename T>
	 struct AudioClip {
		int channels = 0;
		int sampleRate = 0;
		int sampleCount = -1;
		T *data = nullptr;

		AudioClip() = default;
		AudioClip(const AudioClip&) = delete;

		inline ~AudioClip() {
			std::free(data);
			data = nullptr;
		}

		[[nodiscard]]
		inline int size() const {
			return sampleCount * sizeof(T);
		}
	 };

	 using AudioClipShort = AudioClip<short>;
	 using AudioClipByte = AudioClip<char>;

	 constexpr auto ASSET_CLIP_SHORT = assets::make_asset_type("Audio:Clip:Short");
	 constexpr auto ASSET_CLIP_BYTE = assets::make_asset_type("Audio:Clip:Byte");

	 /**
	  * Audio Playback Contract.
	  *
	  * A module depending on this contract my assume that its lifetime is bounded by the existance of a low-level
	  * audio device. In other words, its setup method will be called after the audio device has been created and its
	  * teardown method will be called before the audio device is unavailable.
	  *
	  * A module providing this contract must:
	  *  * Ensure the low-level audio APIs are available once its setup method has been invoked
	  *  * Ensure the low-level audio APIs are cleaned up one its teardown method has been invoked
	  *  * Provide an instance of of an AudioService
	  *
	  */
	 FGGL_MODULE_CONTRACT(AUDIO_PROVIDER, fggl::audio::Provider);

	/**
	 *
	 * \ingroup services
	 */
	class AudioService {
		public:
			FGGL_DEFINE_SERVICE(fggl::audio::AudioService)

			virtual void play(const assets::AssetGUID &asset, bool looping = false) = 0;
			virtual void play(const AudioClipShort &clip, bool looping = false) = 0;

			virtual ~AudioService() = default;
	};

} // namespace fggl::audio

#endif //FGGL_AUDIO_AUDIO_HPP
