/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/06/22.
//

#ifndef FGGL_GFX_INTERFACES2_HPP
#define FGGL_GFX_INTERFACES2_HPP

#include "fggl/gfx/interfaces.hpp"
#include "fggl/display/window.hpp"
#include "fggl/services/module.hpp"

namespace fggl::gfx {

	enum class GraphicsAPI {
		OpenGL, Vulkan
	};

	struct GraphicsDetails {
		GraphicsAPI api;
		int majorVersion;
		int minorVersion;
		bool debug;
	};

	/**
	 * Graphics API Contract.
	 *
	 * This defines the lifetime of any suitable low-level graphics APIs used by the program. A module depending on this
	 * contract may assume that its lifetime will be bounded by that of the graphics API (ie, its setup method will
	 * only be called once the graphics API has been initialised and its teardown method will be called before the
	 * graphics API is shut down).
	 *
	 * A module providing this contract must:
	 *  * Provide services for rendering
	 *  * Initialise required graphics APIs (eg, OpenGL) during its setup method
	 *  * Clean up required graphics APIs (eg, OpenGL) its teardown method
	 *
	 */
	FGGL_MODULE_CONTRACT(API, fggl::gfx::API)

	class WindowGraphics {
		public:
			FGGL_DEFINE_SERVICE(fggl::gfx::WindowGraphics)

			WindowGraphics() = default;
			virtual ~WindowGraphics() = default;

			// no copy
			WindowGraphics(const WindowGraphics& ) = delete;
			WindowGraphics& operator=(const WindowGraphics&) = delete;

			// no move
			WindowGraphics(WindowGraphics&&) = delete;
			WindowGraphics& operator=(WindowGraphics&&) = delete;

			[[nodiscard]]
			virtual GraphicsDetails config() const = 0;

			[[nodiscard]]
			virtual Graphics* create(display::Window &window) = 0;

			inline Graphics* createMain(display::Window& window) {
				assert( m_graphics == nullptr );
				m_graphics = create(window);
				return m_graphics;
			}

			inline void release() {
				if ( m_graphics != nullptr ) {
					delete m_graphics;
					m_graphics = nullptr;
				}
			}

			[[nodiscard]] inline Graphics* get() const {
				return m_graphics;
			}

		private:
			Graphics* m_graphics = nullptr;
	};

} // namespace fggl::gfx

#endif //FGGL_GFX_INTERFACES2_HPP
